# Box Model for polydisperse, particle laden gravity currents
# pybox version 1.0
# April 2, 2019
# Authors: Giovanni Biagioli, Andrea Bevilacqua, Tomaso Esposti Ongaro, Mattia de' Michieli Vitturi
# Istituto Nazionale di Geofisica e Vulcanologia
# Pisa, ITALY

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import scipy
import scipy.linalg   
import scipy.integrate

import pandas as pd
import salem
from salem import DataLevels, GoogleVisibleMap, Map, wgs84
import utm

import sys
import os
import json

from linecache import getline


## Main start here 

# Default input, DEM and output file
inpfile = 'input.dict'
dem_file = 'dem.asc'
outpfile = 'boxmodeloutput'


for arg in sys.argv:
    if(arg[0:2]=='-i'):
        inpfile = arg[2:]
    if(arg[0:2]=='-d'):
        dem_file = arg[2:]
    if(arg[0:2]=='-o'):
        outpfile = arg[2:]

with open(inpfile) as data_file:
    inp_dict = json.load(data_file)
locals().update(inp_dict)

eps0 = np.array(eps0)
rhos = np.array(rhos)
ds = np.array(ds)
nsolid = eps0.size
ninput = len(inp_dict)

#check input number
if flag_DEM:
    if ninput != 24:
        print('Error: missing (or too many) input values. Please correct your input values.')
        sys.exit(-9)
else:
    if ninput != 15 and ninput != 24:
        print('Error: missing (or too many) input values. Please correct your input values.')
        sys.exit(-10)


# Set maximum allowable number of solid phases
Nmax = 21 
if nsolid > Nmax:
    print('Error: number of particle size classes above 21. Please reduce their number.')
    sys.exit(-1)
if rhos.size != nsolid:
    print('Error: inconsistent number of particle densities and volume fractions. Please correct your input values.')
    sys.exit(-2)
if ds.size != nsolid:
    print('Error: inconsistent number of particle sizes and volume fractions. Please correct your input values.')
    sys.exit(-3)
if np.sum(eps0) >= 1:
    print('Error: total volume fraction of solid phase > 1. Please check your volume fraction values.')
    sys.exit(-4)
    
# AUSILIAR SUBROUTINES 

## Subroutine for reading a dem
def read_dem(dem_file, vertical_scale_factor):
    
    # Import .asc file, then read it and convert data in floating-point format
    hdr = [getline(dem_file, i) for i in range(1,7)]
    values = [float(h.split(" ")[-1].strip()) for h in hdr]
    ncols,nrows,xll,yll,cellsize,ndvalue = values
    ncols = int(ncols)
    nrows = int(nrows)
    
    # Compute the UTM coordinates of each element of the topography from the lower-left corner
    xdem = xll+0.5*cellsize + np.linspace(0,(ncols-1)*cellsize, ncols)
    ydem = yll+0.5*cellsize + np.linspace(0,(nrows-1)*cellsize, nrows)
    
    # Load the dem into a numpy array
    zdem = pd.read_csv(dem_file, delim_whitespace=True, header=None,skiprows=6)
    zdem = zdem.apply(pd.to_numeric)
    zdem = np.flipud(zdem)
    zdem = np.maximum(zdem*vertical_scale_factor,0.)
    
    return(ncols, nrows, xll, yll, cellsize, xdem, ydem, zdem)

## Subroutine for writing a dem     
def write_dem(output_file, ncols, nrows, xllcorner, yllcorner, cellsize, elevation):
    fID = open(output_file,'w')
    header = "ncols     %s\n" % ncols
    header += "nrows    %s\n" % nrows
    header += "xllcorner " + str(xllcorner) +"\n"
    header += "yllcorner " + str(yllcorner) +"\n"
    header += "cellsize " + str(cellsize) +"\n"
    header += "NODATA_value 0"
    np.savetxt(output_file, np.flipud(elevation), header=header, fmt='%d', comments='')    
    fID.close()

## Subroutine for dividing a DEM into azimuthal sectors
def dem_section(theta, nmax, x, y, d, ncols, nrows, cellsize, xdem, ydem, zdem, lmax):
    z = np.zeros(nmax+1)
    theta = theta%360
        
    for i in range(1,nmax+1):
        x[i] = x[i-1] + d*np.cos(theta/180*np.pi)
        y[i] = y[i-1] + d*np.sin(theta/180*np.pi)
        if i*d > lmax:
            flag = 2
            i1 = i-1
            break
        elif ((x[i] > xdem[-1]+0.5*cellsize) or (y[i] > ydem[-1]+0.5*cellsize) or (x[i] < xdem[0]-0.5*cellsize) or (y[i] < ydem[0]-0.5*cellsize)):
            flag = 1
            i1 = i-1
            break
        i1 = nmax
        flag = 0
    
    # Find the dem pixel the point (x, y) is in and find the elevation of the point. Then print out the altitude profile
    if theta >= 0. and theta < 90.:
        ii = 0 
        im = ncols 
        istep = 1
        jj = 0 
        jm = nrows
        jstep = 1
    elif theta >= 90. and theta < 180.:
        ii = ncols-1
        im = -1
        istep = -1
        jj = 0
        jm = nrows
        jstep = 1
    elif theta >= 180. and theta < 270.:
        ii = ncols-1
        im = -1
        istep = -1 
        jj = nrows-1
        jm = -1
        jstep = -1
    elif theta >= 270. and theta < 360.:
        ii = 0
        im = ncols
        istep = 1
        jj = nrows-1
        jm = -1
        jstep = -1
    
    for i in range(i1+1):
        for k in range(ii, im, istep):
            if theta >= 0. and theta < 90.:
                cx = x[i] > xdem[k] - 0.5*cellsize
            elif theta >= 90. and theta < 180.:
                cx = x[i] < xdem[k] + 0.5*cellsize
            elif theta >= 180.and theta < 270.:
                cx = x[i] < xdem[k] + 0.5*cellsize
            elif theta >= 270. and theta < 360.:
                cx = x[i] > xdem[k] - 0.5*cellsize
            if cx:
                ii = k
            else:
                break
                
        for t in range(jj, jm, jstep):
            if theta >= 0. and theta < 90.:
                cy = y[i] > ydem[t] - 0.5*cellsize
            elif theta >= 90. and theta < 180.:
                cy = y[i] > ydem[t] - 0.5*cellsize
            elif theta >= 180. and theta < 270.:
                cy = y[i] < ydem[t] + 0.5*cellsize
            elif theta >= 270. and theta < 360.:
                cy = y[i] < ydem[t] + 0.5*cellsize
            if cy:
                jj = t
            else:
                break
        z[i] = zdem[jj,ii]
    return flag,z,i1

## Subroutine for computing the reduced gravity of a polydisperse mixture      
def polydisperse_rg(eps):
    gprime = g*rg
    for i in range(nsolid):
        gprime = gprime + g*phi[i]*eps[i]
    return gprime      

## Subroutine for computing the density of a polydisperse mixture      
def polydisperse_density(eps):
    rhoc = 0.
    epstot = 0.
    for i in range(nsolid):
        rhoc = rhoc + rhos[i]*eps[i]
        epstot = epstot + eps[i]
    rhoc = rhoc + rhog*(1. - epstot)
    return rhoc

def settling(theta, ds, rhos, rhog, kmax, tol):
    mu_g = 1.458e-6*theta**(3./2.)/(theta+110.4)
    ws = ds**2*g*rhos/(18*mu_g)
    ws_old = ws
    Re = rhog*ds*ws/mu_g
    for k in range(kmax):
        C_D = 24.0/Re*(1.+0.15*Re**0.687)
        ws= np.sqrt(4.*rhos*g*ds/(3.*C_D*rhog))
        Re = rhog*ds*ws/mu_g
        if np.abs(ws-ws_old) < tol:
            break
        ws_old = ws
    ws = ws*(Re < 1000.) + (np.sqrt(4.*rhos*g*ds/(3.*1.*rhog)))*(Re >= 1000.)
    return ws
 
## Subroutines for performing numerical integration with RKF45
def fun_cartesian(t, x):
	fun_car = np.zeros(nsolid+1)
	gprime = polydisperse_rg(x[1:])
	if gprime > 0:
	    fun_car[0] = Fr*np.sqrt(gprime*v0/x[0])
	    fun_car[1:] = -ws[:]*x[1:]*x[0]/v0
	return fun_car
	
def fun_cylindrical(t, x):
	fun_cyl = np.zeros(nsolid+1)
	gprime = polydisperse_rg(x[1:])
	if gprime > 0:
		fun_cyl[0] = Fr*np.sqrt(gprime*v0/(f*np.pi))/x[0]
		fun_cyl[1:] = -np.pi*f*ws[:]*x[1:]*(x[0]**2)/v0
	return fun_cyl
	
def event_gip(t,x):
	gip = polydisperse_rg(x[1:])
	return gip

## Main function
def boxModel(dem_file, outpfile, flag_DEM, flag_coords, dt, l0, h0, theta0, nsolid, eps0, rhos, ds, alpha, Fr, g, rhoa, r_tol, a_tol):
    
    global rhog
    rhog = rhoa*300/theta0
    global phi
    phi = np.zeros(nsolid)
    global rg
    rg = (rhog - rhoa)/rhoa
    phi[:] = (rhos[:] - rhog)/rhoa
    rhoc = polydisperse_density(eps0[:])
    if rhoc<rhoa:
        print('Initial density of the current lower than density of surrounding ambient fluid')
        sys.exit(-5) 
    
    global ws
    ws = np.zeros(nsolid)
    for i in range(nsolid):
        ws[i] = settling(theta0, ds[i], rhos[i], rhog, 100, 1e-4)
    if int(theta0) != 300:
        tmax = (np.log((rhoa-rhog)/np.sum(eps0*(rhos-rhog)))*(-h0/np.min(ws)))
    else:
        tmax = (np.log((np.sum(eps0)*rhoa)/np.sum(eps0*rhos)))*(-h0/np.min(ws))
    nt = int(tmax/dt)+1
    
    # Read the DEM if prescribed
    if flag_DEM:
        if anglemin%360. == anglemax:
            print('Error: null angle. Please modify anglemin or anglemax.')
            sys.exit(-6)
        if anglemin<0 or anglemax>360:
            print('Error: invalid angle. Please input anglemin > 0, and anglemax < 360.')
            sys.exit(-7)
        [nx, ny, xll, yll, cellsize, xdem, ydem, zdem] = read_dem(dem_file, vertical_scale_factor)
        if (xv < xdem[0]-0.5*cellsize or xv > xdem[-1] + 0.5*cellsize) or (yv < ydem[0]-0.5*cellsize or yv > ydem[-1] + 0.5*cellsize):
            print('Error: the vent location is outside DEM range. Please provide xv in: [',xdem[0],',',xdem[-1]+cellsize,'] and yv in [',ydem[0],',',ydem[-1]+cellsize,']')
            sys.exit(-8) 
                
    # Initialize arrays with the box model solution (front positions, depths, times, velocities of the current front, solid phases concentrations, solid phases deposits)
    l = np.zeros(nt)
    h = np.zeros(nt)
    t = np.zeros(nt)
    u = np.zeros(nt)
    tke = np.zeros(nt)
    tpe = np.zeros(nt)
    hmax = np.zeros(nt)
    eps = np.zeros((nt, nsolid))
    deposit = np.zeros((nt, nsolid))
    
    # Set the initial state
    time = 0.
    eps[0,:] = eps0[:]
    l[0] = l0
    h[0] = h0
    deposit[0,:] = np.zeros(nsolid)
    gprime = polydisperse_rg(eps[0,:])
    u[0] = Fr*np.sqrt(gprime*h0)        
    tke[0] = 0.5*rhoc*(u[0]**2)
    hmax[0] = 1e4
    global v0
    v0 = 0.
    
    ## Write box model solution on a proper file
    fID = open(outpfile + '.dat', 'w')
    fID.write('time (s)                                                         granulometric class concentration                                                                                height (m)     length (m)   rho_c (kg/m^3)     u (m/s)       potential energy (J)    kinetic energy (J)      hmax (m) \n \n')
    
    y0 = np.zeros(nsolid+1)
    y0[0] = l0
    for i in range(nsolid):
        y0[i+1] = eps0[i]
    event_gip.terminal = True
    event_gip.direction = -1
    
    # Solve box model equations in Cartesian coordinates
    if flag_coords > 0:
    	v0 = l0*h0
    	tpe[0] = gprime*rhoa*v0*h[0]
    	
    	fID.write(' %3.2f          ' % time)
    	for n in range(nsolid):
    		fID.write("%f   " % eps[0,n])
    	fID.write('       %5.3f       %5.3f         %5.3f           %5.3e          %5.3e              %3.3e            %3.3e \n' % (h0, l0, rhoc, u[0], tpe[0], tke[0], hmax[0]))
    	
    	# Numerical integration of differential equations with RKF45 method		
    	sol = scipy.integrate.solve_ivp(fun_cartesian, [0, tmax], y0, method = 'RK45', t_eval = np.arange(0, tmax, dt), events = event_gip, rtol = r_tol, atol = a_tol)
    	t = sol.t
    	y = sol.y
    	for i in range(1,len(t)):
        	time = t[i]
        	l[i] = y[0,i]
        	eps[i,:] = y[1:,i]
        	gip = polydisperse_rg(eps[i,:])
        	if gip < 0:
        	    break
        	deposit[i,:] = ws[:]*rhos[:]*eps[i,:]*dt
        	l_max = l[i]
        	h[i] = v0/l[i]
        	u[i] = Fr*np.sqrt(gip*h[i])
        	tpe[i] = gip*rhoa*v0*h[i]
        	rhom = polydisperse_density(eps[i,:])
        	tke[i] = 0.5*rhom*(u[i]**2)
        	hmax[i] = 0.5*(1./g)*(rhom/(rhom - rhoa))*(u[i]**2)
        	i_max = i
        	fID.write(' %3.2f          ' % time)
        	for n in range(nsolid):
        		fID.write("%f   " % eps[i,n])
        	fID.write('       %5.3f       %5.3f         %5.3f           %5.3e          %5.3e              %3.3e            %3.3e \n' %(h[i], l[i], rhom, u[i], tpe[i], tke[i], hmax[i]))
         
    # Solve box model equations in Cylindrical coordinates 
    else:
        if  flag_DEM==0:
            spreading = 360
            sr = 25
        elif anglemax > anglemin:
            spreading = anglemax - anglemin
        else:
            spreading = anglemax + (360. - anglemin)
        global f
        f = spreading/360.
        v0 = np.pi*(l0**2)*h0*f
        tpe[0] = gprime*rhoa*v0*h[0]
        
        fID.write(' %3.2f          ' % time)
        for n in range(nsolid):
            fID.write("%f   " % eps[0,n])
        fID.write('       %5.3f       %5.3f         %5.3f           %5.3e          %5.3e              %3.3e            %3.3e \n' % (h0, l0, rhoc, u[0], tpe[0], tke[0], hmax[0]))
        
        sol = scipy.integrate.solve_ivp(fun_cylindrical, [0, tmax], y0, method = 'RK45', t_eval = np.arange(0, tmax, dt), events = event_gip, rtol = r_tol, atol = a_tol)
        t = sol.t
        y = sol.y
        for i in range(1,len(t)):
            time = t[i]
            l[i] = y[0,i]
            eps[i,:] = y[1:,i]
            gip = polydisperse_rg(eps[i,:])
            if gip < 0:
                break
            deposit[i,:] = ws[:]*rhos[:]*eps[i,:]*dt
            l_max = l[i]
            h[i] = v0/((l[i]**2)*np.pi*f)
            u[i] = Fr*np.sqrt(gip*h[i])
            tpe[i] = gip*rhoa*v0*h[i]
            rhom = polydisperse_density(eps[i,:])
            tke[i] = 0.5*rhom*(u[i]**2)
            hmax[i] = 0.5*(1./g)*(rhom/(rhom - rhoa))*(u[i]**2)
            i_max = i
            fID.write(' %3.2f          ' % time)
            for n in range(nsolid):
                fID.write("%f   " % eps[i,n])
            fID.write('       %5.3f       %5.3f         %5.3f           %5.3e          %5.3e              %3.3e            %3.3e \n' %(h[i], l[i], rhom, u[i], tpe[i], tke[i], hmax[i]))
    fID.close()
    
    # Create the array hmax(l)
    if flag_DEM:
       sr = rad_res
    lmax_int = int(l_max+sr)
    hmax_of_l = np.zeros(lmax_int+1)
    hmax_of_l[0] = hmax[0]
    for i in range(1, i_max+1):
        lint = int(l[i])
        hmax_of_l[lint] = hmax[i]
    for k in range(1,int(l_max)+1):
        if (hmax_of_l[k] == 0.):
            hmax_of_l[k]=hmax_of_l[k-1]
    
    # Write the deposit profiles: deposited total mass of each particle class, deposited mass fraction of each particle class, deposit thickness
    massd = np.zeros(nsolid)
    fract = np.zeros(nsolid)
    thick = np.zeros(nsolid)
    fID2 = open(outpfile + '_thickness.dat', 'w')
    fID2.write('current front position (m)                                                                                            granulometric class deposit thickness (m)                                                                                          global deposit thickness (m)\n')
    for i in range(1,i_max+1):
        totalmass = np.sum(deposit[i:i_max+1,:])
        massd[:] = np.sum(deposit[i:i_max+1,:])
        fract[:] = np.sum(deposit[i:i_max+1,:])/totalmass
        thick[:] = massd[:]/(alpha*rhos[:])
        fID2.write('        %5.3e                ' % l[i])
        for n in range(nsolid):
            fID2.write('%3.3e         ' % thick[n])
        fID2.write('        %3.3e \n' % np.sum(thick[:]))
    fID2.close()

    if flag_DEM:
        if flag_coords < 0:
            Xdem,Ydem = np.meshgrid(xdem,ydem)
            Dist = np.sqrt((Xdem-xv)**2 + (Ydem-yv)**2)
                    
            # Energy-conoid: compare topography matrix zdem and hmax along sectors.
            nrmax = int(np.ceil(np.sqrt(nx**2+ny**2)*cellsize/sr))
            xr = np.zeros(nrmax+1)
            yr = np.zeros(nrmax+1)
            xr[0] = xv
            yr[0] = yv
            
            # Loop over azimuth angle and store the maximum distance dmax(angle)
            number_of_sectors = 360
            dmax = np.zeros(number_of_sectors)
            dbeta = 360./number_of_sectors
            if anglemin < anglemax:
                if (anglemin == 0. and anglemax >= 359.):
                    angles = range(number_of_sectors)
                elif (anglemin > 0. and  anglemax >= 359.):
                    angles = np.unique(np.concatenate((np.arange(anglemin - 1., 360.), np.array([0.]))))
                elif (anglemin == 0. and anglemax < 359.):
                    angles = np.unique(np.concatenate((np.array([359.]), np.arange(0., anglemax+2.))))
                elif (anglemin > 0. and anglemax < 359.):
                    angles = np.arange(anglemin-1., anglemax + 2.)
            else:
                angles = np.concatenate((np.arange(anglemin-1., 360.), np.arange(0, anglemax+2.)))

            for nb in angles:
                flag, zr, imax = dem_section(nb, nrmax, xr, yr, sr, nx, ny, cellsize, xdem, ydem, zdem, lmax_int)
                zrmin = zr[0]
                if flag == 1:
                    imax = imax+2
                if flag == 2:
                    imax = imax+1
                for na in range(1,imax):
                    dist = int(sr*na)
                    hmx = hmax_of_l[dist]
                    dmax[int(nb)] = sr*na
                    
                    # Energy conoid condition
                    if differential_topography:
                        zrmin = np.min((zr[na], zrmin))
                        if (zr[na] - zrmin > hmx):
                            break
                    else:
                        if (zr[na] > hmx):
                            break
                            
            # Build the invasion map
            invasion = np.zeros((ny, nx))
            Angle = np.arctan2(Ydem - yv, Xdem - xv)
            Angle = Angle*180./np.pi
            Angle = Angle%360
            if anglemin < anglemax:
                insector = (Angle > anglemin)*(Angle < anglemax)
            elif anglemin > anglemax:
                insector = (Angle > anglemin)+(Angle < anglemax)
            nb1 = (np.ceil(Angle)%360).astype(int)
            nb2 = Angle.astype(int)
            invasion = ((Dist < np.maximum(dmax[nb1], dmax[nb2]))*insector).astype(int)
            
            # Write invasion map in ESRI ASCII format
            output_file = outpfile + '_EC2.asc'
            write_dem(output_file, nx, ny, xll, yll, cellsize, invasion)
                                   
            ## Visualize invasion map 
            
            # Convert UTM coordinates to latitude-longitude
            demMin = utm.to_latlon(xll, yll, zone_number, zone_letter)
            demMax = utm.to_latlon(xdem[-1]+0.5*cellsize, ydem[-1]+0.5*cellsize, zone_number, zone_letter)
            vent_coords =  utm.to_latlon(xv, yv, zone_number, zone_letter)
            min_x = demMin[1]
            min_y = demMin[0]
            max_x = demMax[1]
            max_y = demMax[0]

            f = plt.figure()
            plt.rcParams["font.size"] = 8.0
            cm = ListedColormap(['gold', 'midnightblue'])
            g = GoogleVisibleMap(x=[min_x, max_x], y=[min_y, max_y], maptype='terrain')  
            sm = Map(g.grid, factor=1, countries=False)
            ggl_img = g.get_vardata()
            sm.set_rgb(ggl_img)
            sm.visualize()
            a = (max_x - min_x)/nx
            b = (max_y - min_y)/ny            
            grid = salem.Grid(nxny=(nx+1, ny+1), dxdy = (a,b), x0y0=(min_x, min_y))
            lon, lat = grid.ll_coordinates		

            # convert Lat, Lon to coordinates for figure
            x, y = sm.grid.transform(lon, lat)

            im = plt.pcolormesh(x,y,invasion, cmap=cm, alpha = 0.05)
            cbar = f.colorbar(im, ticks=np.linspace(0,1,2))
            cbar.solids.set(alpha=1)
            plt.grid()
            plt.tight_layout()

            # Insert the vent
            ivent = np.argmin(np.absolute(xdem - 0.5*cellsize - xv),axis=0)
            jvent = np.argmin(np.absolute(ydem - 0.5*cellsize - yv),axis=0)
            xv_ll = x[1,ivent]
            yv_ll = y[jvent,1]	
            plt.scatter(xv_ll, yv_ll, color='red', s = 3, edgecolors='red')
            plt.legend(['vent'], scatterpoints=1, loc='lower right', fontsize=8)
            f.savefig(outpfile + '.png',dpi=300)
 
boxModel(dem_file, outpfile, flag_DEM, flag_coords, dt, l0, h0, theta0, nsolid, eps0, rhos, ds, alpha, Fr, g, rhoa, r_tol, a_tol)
